#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import binascii
from collections import namedtuple
from platform import python_version

from six import ensure_binary, ensure_str

from ._version import __version__

# Sentinel value used to highlight default values
# when 'None' has special meaning (like request_timeout)
# The only supported operation on this value is identity ('is DEFAULT')
DEFAULT = namedtuple("DEFAULT", ())()


def create_user_agent(name, version):
    """Creates the 'User-Agent' header given the library name and version"""
    return "%s/%s (Python/%s; elastic-transport/%s)" % (
        name,
        version,
        python_version(),
        __version__,
    )


def normalize_headers(headers):
    """Normalizes HTTP headers to be lowercase to ensure
    there are no case-collisions deeper within the stack.
    """
    if not headers:
        return {}
    return {
        k.lower(): v
        for k, v in (headers.items() if hasattr(headers, "items") else headers)
    }


CloudID = namedtuple(
    "CloudID", ["cluster_name", "es_host", "es_port", "kibana_host", "kibana_port"]
)


def parse_cloud_id(cloud_id):
    """Parses a Cloud ID into its components"""
    try:
        cloud_id = ensure_str(cloud_id)
        cluster_name, _, cloud_id = cloud_id.partition(":")
        parts = ensure_str(
            binascii.a2b_base64(ensure_binary(cloud_id, "ascii")), "ascii"
        ).split("$")
        parent_dn = parts[0]
        if not parent_dn:
            raise ValueError()  # Caught and re-raised properly below
        try:
            es_uuid = parts[1]
        except IndexError:
            es_uuid = None
        try:
            kibana_uuid = parts[2] or None
        except IndexError:
            kibana_uuid = None
        port = None
        if ":" in parent_dn:
            parent_dn, _, parent_port = parent_dn.rpartition(":")
            if parent_port != "443":
                port = int(parent_port)
    except (ValueError, IndexError, UnicodeError):
        raise ValueError("Cloud ID is not properly formatted")

    es_host = "%s.%s" % (es_uuid, parent_dn) if es_uuid else None
    kibana_host = "%s.%s" % (kibana_uuid, parent_dn) if kibana_uuid else None

    return CloudID(
        cluster_name=cluster_name,
        es_host=es_host,
        es_port=port,
        kibana_host=kibana_host,
        kibana_port=port,
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.


class Response(object):
    """HTTP response"""

    def __init__(self, headers, status, body):
        self.headers = headers
        self.status = status
        self.body = body

    def __repr__(self):
        return repr(self.body)

    def __str__(self):
        return str(self.body)

    def __getattr__(self, item):
        return getattr(self.body, item)

    def __getitem__(self, item):
        return self.body[item]

    def __bool__(self):
        return bool(self.body)

    # Python 2 compatibility
    __nonzero__ = __bool__

    def __iter__(self):
        return iter(self.body)

    def __contains__(self, item):
        return item in self.body

    def __len__(self):
        return len(self.body)

    def __eq__(self, other):
        if isinstance(other, type(self.body)):
            return other == self.body
        elif isinstance(other, Response):
            return other.body == self.body
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, type(self.body)):
            return other != self.body
        elif isinstance(other, Response):
            return other.body != self.body
        return NotImplemented


class DictResponse(Response, dict):
    def __init__(self, headers, status, body):
        Response.__init__(self, headers, status, body)
        dict.__init__(self, body)


class ListResponse(Response, list):
    def __init__(self, headers, status, body):
        Response.__init__(self, headers, status, body)
        list.__init__(self, body)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

"""Transport classes and utilities shared among Python Elastic client libraries"""

from ._version import __version__  # noqa
from .connection import Connection, RequestsHttpConnection, Urllib3HttpConnection
from .connection_pool import (
    ConnectionPool,
    ConnectionSelector,
    DummyConnectionPool,
    EmptyConnectionPool,
    RandomSelector,
    RoundRobinSelector,
)
from .exceptions import (
    APIError,
    BadGatewayError,
    BadRequestError,
    ConflictError,
    ConnectionError,
    ConnectionTimeout,
    ForbiddenError,
    GatewayTimeoutError,
    InternalServerError,
    MethodNotImplementedError,
    NotFoundError,
    PayloadTooLargeError,
    PaymentRequiredError,
    SerializationError,
    ServiceUnavailableError,
    TransportError,
    UnauthorizedError,
    UnprocessableEntityError,
)
from .serializer import Deserializer, JSONSerializer, Serializer, TextSerializer
from .transport import Transport

__all__ = [
    "APIError",
    "BadGatewayError",
    "BadRequestError",
    "ConflictError",
    "Connection",
    "ConnectionError",
    "ConnectionPool",
    "ConnectionSelector",
    "ConnectionTimeout",
    "Deserializer",
    "DummyConnectionPool",
    "EmptyConnectionPool",
    "ForbiddenError",
    "GatewayTimeoutError",
    "InternalServerError",
    "JSONSerializer",
    "MethodNotImplementedError",
    "NotFoundError",
    "PayloadTooLargeError",
    "PaymentRequiredError",
    "RandomSelector",
    "RequestsHttpConnection",
    "RoundRobinSelector",
    "SerializationError",
    "Serializer",
    "ServiceUnavailableError",
    "TextSerializer",
    "Transport",
    "TransportError",
    "UnauthorizedError",
    "UnprocessableEntityError",
    "Urllib3HttpConnection",
]



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from six import add_metaclass, python_2_unicode_compatible

HTTP_EXCEPTIONS = {}


class TransportErrorMeta(type):
    def __new__(meta_cls, *args, **kwargs):
        cls = type.__new__(meta_cls, *args, **kwargs)
        status = getattr(cls, "status", None)
        if status is not None:
            HTTP_EXCEPTIONS[status] = cls
        return cls


@python_2_unicode_compatible
@add_metaclass(TransportErrorMeta)
class TransportError(Exception):
    """Generic exception for the 'elastic-transport' package.

    For the 'errors' attribute, errors are ordered from
    most recently raised (index=0) to least recently raised (index=N)

    If an HTTP status code is available with the error it
    will be stored under 'status'.
    """

    status = None

    def __init__(self, message, errors=(), status=None):
        super(TransportError, self).__init__(message)
        self.errors = tuple(errors)
        self.message = message
        if status is not None:
            self.status = status

    def __repr__(self):
        parts = [repr(self.message)]
        if self.status is not None:
            parts.append("status=%r" % self.status)
        if self.errors:
            parts.append("errors=%r" % (self.errors,))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(parts))

    def __str__(self):
        if self.status:
            return "[%s] %s" % (self.status, self.message)
        return str(self.message)


class SerializationError(TransportError):
    """Error that occurred during the serialization or
    deserialization of an HTTP message body
    """


class ConnectionError(TransportError):
    """Error raised by the HTTP connection"""


class ConnectionTimeout(TransportError):
    """Connection timed out during an operation"""


class APIError(TransportError):
    """Error that is raised from the service via HTTP status codes"""


class BadRequestError(APIError):
    """Error for HTTP status 400 'Bad Request'"""

    status = 400


class UnauthorizedError(APIError):
    """Error for HTTP status 401 'Unauthorized'"""

    status = 401


class PaymentRequiredError(APIError):
    """Error for HTTP status 402 'Payment Required'
    Usually signals that your instance doesn't have
    a proper license active for the operation
    """

    status = 402


class ForbiddenError(APIError):
    """Error for HTTP status 403 'Forbidden'"""

    status = 403


class NotFoundError(APIError):
    """Error for HTTP status 404 'Not Found'"""

    status = 404


class ConflictError(APIError):
    """Error for HTTP status 409 'Conflict'"""

    status = 409


class PayloadTooLargeError(APIError):
    """Error for HTTP status 413 'Payload Too Large'"""

    status = 413


class UnprocessableEntityError(APIError):
    """Error for HTTP status 422 'Unprocessable Entity'"""

    status = 422


class InternalServerError(APIError):
    """Error for HTTP status 500 'Internal Server Error'"""

    status = 500


class MethodNotImplementedError(APIError):
    """Error for HTTP status 501 'Method Not Allowed'"""

    status = 501


class BadGatewayError(APIError):
    """Error for HTTP status 502 'Bad Gateway'"""

    status = 502


class ServiceUnavailableError(APIError):
    """Error for HTTP status 503 'Service Unavailable'"""

    status = 503


class GatewayTimeoutError(APIError):
    """Error for HTTP status 504 'Gateway Timeout'"""

    status = 504



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

try:
    import simplejson as json
except ImportError:
    import json

import uuid
from datetime import date
from decimal import Decimal

import six

from .compat import string_types
from .exceptions import SerializationError


class Serializer(object):
    mimetype = None

    def loads(self, s):  # pragma: nocover
        raise NotImplementedError()

    def dumps(self, data):  # pragma: nocover
        raise NotImplementedError()


class TextSerializer(Serializer):
    mimetype = "text/plain"

    def loads(self, s):
        return s

    def dumps(self, data):
        if isinstance(data, string_types):
            return data
        raise SerializationError("Cannot serialize %r into text" % data)


class JSONSerializer(Serializer):
    mimetype = "application/json"

    def default(self, data):
        if isinstance(data, date):
            return data.isoformat()
        elif isinstance(data, uuid.UUID):
            return str(data)
        elif isinstance(data, Decimal):
            return float(data)
        raise SerializationError(
            "Unable to serialize %r (type: %s)" % (data, type(data))
        )

    def loads(self, s):
        try:
            return json.loads(s)
        except (ValueError, TypeError) as e:
            raise SerializationError(message=s, errors=(e,))

    def dumps(self, data):
        if isinstance(data, string_types):
            return data
        try:
            return json.dumps(
                data, default=self.default, ensure_ascii=False, separators=(",", ":")
            )
        # This should be captured by the .default()
        # call but just in case we also wrap these.
        except (ValueError, UnicodeError, TypeError) as e:  # pragma: nocover
            raise SerializationError(
                message="Unable to serialize %r (type: %s)" % (data, type(data)),
                errors=(e,),
            )


DEFAULT_SERIALIZERS = {
    JSONSerializer.mimetype: JSONSerializer(),
    TextSerializer.mimetype: TextSerializer(),
}


class Deserializer(object):
    def __init__(self, serializers, default_mimetype="application/json"):
        try:
            self.default = serializers[default_mimetype]
        except KeyError:
            raise ValueError("Cannot find default serializer (%s)" % default_mimetype)
        self.serializers = serializers

    def loads(self, s, mimetype=None):
        if not mimetype:
            deserializer = self.default
        else:
            # split out charset
            mimetype, _, _ = mimetype.partition(";")
            try:
                deserializer = self.serializers[mimetype]
            except KeyError:
                return six.raise_from(
                    SerializationError(
                        "Unknown mimetype, unable to deserialize: %s" % mimetype
                    ),
                    None,
                )

        return deserializer.loads(s)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import io
import logging

import six

from ..exceptions import HTTP_EXCEPTIONS, APIError
from ..utils import DEFAULT, normalize_headers

try:
    import simplejson as json
except ImportError:
    import json

logger = logging.getLogger("elastic_transport.connection")


class Connection(object):
    """
    Class responsible for maintaining a connection to an Enterprise Search node. It
    holds persistent connection pool to it and it's main interface
    (`perform_request`) is thread-safe.

    Also responsible for logging.

    :arg host: hostname of the node (default: localhost)
    :arg port: port to use (integer, default: 9200)
    :arg use_ssl: use ssl for the connection if `True`
    :arg path_prefix: optional url prefix for Enterprise Search
    :arg timeout: default timeout in seconds (float, default: 10)
    :arg http_compress: Use gzip compression
    :arg opaque_id: Send this value in the 'X-Opaque-Id' HTTP header
        For tracing all requests made by this transport.
    :arg user_agent: 'User-Agent' HTTP header for the given service.
    """

    def __init__(
        self,
        host="localhost",
        port=None,
        use_ssl=False,
        url_prefix="",
        request_timeout=10,
        headers=None,
        http_compress=None,
        opaque_id=None,
        user_agent=None,
        **kwargs
    ):
        # Work-around if the implementing class doesn't
        # define the headers property before calling super().__init__()
        if not hasattr(self, "headers"):
            self.headers = {}

        self.headers.update(normalize_headers(headers))
        if opaque_id:
            self.headers["x-opaque-id"] = opaque_id
        if user_agent:
            self.headers.setdefault("user-agent", user_agent)

        self.headers.setdefault("content-type", "application/json")
        if http_compress:
            self.headers["accept-encoding"] = "gzip"

        scheme = kwargs.get("scheme", "http")
        if use_ssl or scheme == "https":
            scheme = "https"
            use_ssl = True
        self.use_ssl = use_ssl
        self.http_compress = http_compress or False

        self.scheme = scheme
        self.port = port
        self.host = host
        if url_prefix:
            url_prefix = "/" + url_prefix.strip("/")
        self.url_prefix = url_prefix
        self.request_timeout = request_timeout

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.base_url)

    def __eq__(self, other):
        if not isinstance(other, Connection):
            raise TypeError("Unsupported equality check for %s and %s" % (self, other))
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return id(self)

    @property
    def base_url(self):
        return "".join(
            [
                self.scheme,
                "://",
                # IPv6 must be wrapped by [...]
                "[%s]" % self.host if ":" in self.host else self.host,
                ":%s" % self.port if self.port is not None else "",
                self.url_prefix,
            ]
        )

    def perform_request(
        self,
        method,
        path,
        params=None,
        body=None,
        request_timeout=DEFAULT,
        ignore_status=(),
        headers=None,
    ):  # pragma: nocover
        raise NotImplementedError()

    def log_request_success(self, method, url, body, status, response, duration):
        """Log a successful API call"""
        # body has already been serialized to utf-8, deserialize it for logging
        # TODO: find a better way to avoid (de)encoding the body back and forth
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.info("%s %s [status:%s request:%.3fs]", method, url, status, duration)
        logger.debug("> %s", body)
        logger.debug("< %s", response)

    def log_request_fail(
        self,
        method,
        url,
        body,
        duration,
        status=None,
        response=None,
        exception=None,
    ):
        """Log an unsuccessful API call"""
        # do not log 404s on HEAD requests
        if method == "HEAD" and status == 404:
            return
        logger.warning(
            "%s %s [status:%s request:%.3fs]",
            method,
            url,
            status or "N/A",
            duration,
            exc_info=exception is not None,
        )

        # body has already been serialized to utf-8, deserialize it for logging
        # TODO: find a better way to avoid (de)encoding the body back and forth
        if body:
            try:
                body = body.decode("utf-8", "ignore")
            except AttributeError:
                pass

        logger.debug("> %s", body)

        if response is not None:
            logger.debug("< %s", response)

    def _raise_error(self, status, raw_data):
        """Locate appropriate exception and raise it. Attempts
        to decode the raw data as JSON for better usability.
        """
        try:
            raw_data = json.loads(six.ensure_str(raw_data, "utf-8", "ignore"))
        except Exception:
            pass
        raise HTTP_EXCEPTIONS.get(status, APIError)(message=raw_data, status=status)

    def _gzip_compress(self, body):
        buf = io.BytesIO()
        with gzip.GzipFile(fileobj=buf, mode="wb") as f:
            f.write(body)
        return buf.getvalue()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from .compat import string_types, urlparse
from .connection import RequestsHttpConnection, Urllib3HttpConnection
from .connection_pool import ConnectionPool, DummyConnectionPool, EmptyConnectionPool
from .exceptions import ConnectionError, ConnectionTimeout, TransportError
from .response import DictResponse, ListResponse, Response
from .serializer import DEFAULT_SERIALIZERS, Deserializer
from .utils import DEFAULT, normalize_headers

# Allows for using a connection_class by name rather than import.
CONNECTION_CLASS_NAMES = {
    "urllib3": Urllib3HttpConnection,
    "requests": RequestsHttpConnection,
}


class Transport(object):
    """
    Encapsulation of transport-related to logic. Handles instantiation of the
    individual connections as well as creating a connection pool to hold them.

    Main interface is the `perform_request` method.
    """

    DEFAULT_CONNECTION_CLASS = Urllib3HttpConnection

    def __init__(
        self,
        hosts=None,
        connection_class=None,
        connection_pool_class=ConnectionPool,
        serializers=None,
        default_mimetype="application/json",
        default_hosts=None,
        max_retries=3,
        retry_on_status=(502, 503, 504),
        retry_on_timeout=False,
        **kwargs
    ):
        """
        :arg hosts: list of dictionaries, each containing keyword arguments to
            create a `connection_class` instance
        :arg connection_class: subclass of :class:`~elastic_transport.Connection` to use
            or the name of the Connection (ie 'urllib3', 'requests')
        :arg connection_pool_class: subclass of :class:`~elastic_transport.ConnectionPool` to use
        :arg serializers: optional dict of serializer instances that will be
            used for deserializing data coming from the server. (key is the mimetype)
        :arg default_mimetype: when no mimetype is specified by the server
            response assume this mimetype, defaults to `'application/json'`
        :arg default_hosts: Default hosts config to use if none is given.
        :arg max_retries: maximum number of retries before an exception is propagated
        :arg retry_on_status: set of HTTP status codes on which we should retry
            on a different node. defaults to ``(502, 503, 504)``
        :arg retry_on_timeout: should timeout trigger a retry on different
            node? (default ``False``)

        Any extra keyword arguments will be passed to the `connection_class`
        when creating and instance unless overridden by that connection's
        options provided as part of the hosts parameter.
        """
        hosts = _normalize_hosts(hosts, default_hosts)
        if connection_class is None:
            connection_class = self.DEFAULT_CONNECTION_CLASS
        elif isinstance(connection_class, str):
            if connection_class not in CONNECTION_CLASS_NAMES:
                raise ValueError(
                    "Unknown option for connection_class: '%s'. "
                    "Available options are: '%s'"
                    % (
                        connection_class,
                        "', '".join(sorted(CONNECTION_CLASS_NAMES.keys())),
                    )
                )
            connection_class = CONNECTION_CLASS_NAMES[connection_class]

        # serialization config
        _serializers = DEFAULT_SERIALIZERS.copy()
        # if custom serializers map has been supplied, override the defaults with it
        if serializers:
            _serializers.update(serializers)
        # create a deserializer with our config
        self.deserializer = Deserializer(_serializers, default_mimetype)

        self.max_retries = max_retries
        self.retry_on_timeout = retry_on_timeout
        self.retry_on_status = retry_on_status

        # data serializer
        self.serializer = _serializers[default_mimetype]

        # store all strategies...
        self.connection_pool_class = connection_pool_class
        self.connection_class = connection_class

        # ...save kwargs to be passed to the connections
        self.kwargs = kwargs
        self.hosts = hosts

        # Start with an empty pool specifically for `AsyncTransport`.
        # It should never be used, will be replaced on first call to
        # .set_connections()
        self.connection_pool = EmptyConnectionPool()

        if hosts:
            # ...and instantiate them
            self.set_connections(hosts)
            # retain the original connection instances for sniffing
            self.seed_connections = list(self.connection_pool.connections[:])
        else:
            self.seed_connections = []

    def add_connection(self, host):
        """
        Create a new :class:`~elastic_enterprise_search.Connection` instance and add it to the pool.

        :arg host: kwargs that will be used to create the instance
        """
        self.hosts.append(host)
        self.set_connections(self.hosts)

    def set_connections(self, hosts):
        """
        Instantiate all the connections and create new connection pool to hold them.
        Tries to identify unchanged hosts and re-use existing
        :class:`~elastic_transport.Connection` instances.

        :arg hosts: same as `__init__`
        """
        # construct the connections
        def _create_connection(host):
            # if this is not the initial setup look at the existing connection
            # options and identify connections that haven't changed and can be
            # kept around.
            if hasattr(self, "connection_pool"):
                for (connection, old_host) in self.connection_pool.connection_opts:
                    if old_host == host:
                        return connection

            # previously unseen params, create new connection
            kwargs = self.kwargs.copy()
            kwargs.update(host)
            return self.connection_class(**kwargs)

        connections = map(_create_connection, hosts)

        connections = list(zip(connections, hosts))
        if len(connections) == 1:
            self.connection_pool = DummyConnectionPool(connections)
        else:
            # pass the hosts dicts to the connection pool to optionally extract parameters from
            self.connection_pool = self.connection_pool_class(
                connections, **self.kwargs
            )

    def get_connection(self):
        """
        Retrieve a :class:`~elastic_transport.Connection` instance from the
        :class:`~elastic_transport.ConnectionPool` instance.
        """
        return self.connection_pool.get_connection()

    def mark_dead(self, connection):
        """
        Mark a connection as dead (failed) in the connection pool. If sniffing
        on failure is enabled this will initiate the sniffing process.

        :arg connection: instance of :class:`~elastic_transport.Connection` that failed
        """
        # mark as dead even when sniffing to avoid hitting this host during the sniff process
        self.connection_pool.mark_dead(connection)

    def perform_request(
        self,
        method,
        path,
        headers=None,
        params=None,
        body=None,
        request_timeout=DEFAULT,
        ignore_status=(),
    ):
        """
        Perform the actual request. Retrieve a connection from the connection
        pool, pass all the information to it's perform_request method and
        return the data.

        If an exception was raised, mark the connection as failed and retry (up
        to `max_retries` times).

        If the operation was successful and the connection used was previously
        marked as dead, mark it as live, resetting it's failure count.

        :arg method: HTTP method to use
        :arg path: relative URL to target
        :arg headers: dictionary of headers, will be handed over to the
            underlying :class:`~elastic_transport.Connection` class
        :arg params: dictionary of query parameters, will be handed over to the
            underlying :class:`~elastic_transport.Connection` class for serialization
        :arg body: body of the request, will be serialized using serializer and
            passed to the connection
        :arg request_timeout: Timeout to be passed to the HTTP client for the request
        :arg ignore_status: Collection of HTTP status codes to not raise an error for.
        :returns: Deserialized Response
        """
        if isinstance(ignore_status, int):
            ignore_status = (ignore_status,)

        if body is not None:
            body = self.serializer.dumps(body)

        if body is not None:
            try:
                body = body.encode("utf-8", "surrogatepass")
            except (UnicodeError, AttributeError):
                # bytes/str - no need to re-encode
                pass

        headers = normalize_headers(headers)

        # Errors are stored from (oldest->newest)
        errors = []

        for attempt in range(self.max_retries + 1):
            connection = self.get_connection()

            try:
                status, headers_response, data = connection.perform_request(
                    method,
                    path,
                    params,
                    body,
                    headers=headers,
                    ignore_status=ignore_status,
                    request_timeout=request_timeout,
                )
            except TransportError as e:
                if method == "HEAD" and e.status == 404:
                    return Response(
                        status=404,
                        headers={},
                        body=False,
                    )

                retry = False
                if isinstance(e, ConnectionTimeout):
                    retry = self.retry_on_timeout
                elif isinstance(e, ConnectionError):
                    retry = True
                elif e.status in self.retry_on_status:
                    retry = True

                if retry:
                    try:
                        # only mark as dead if we are retrying
                        self.mark_dead(connection)
                    except TransportError:
                        # If sniffing on failure, it could fail too. Catch the
                        # exception not to interrupt the retries.
                        pass
                    # raise exception on last retry
                    if attempt == self.max_retries:
                        e.errors = tuple(errors)
                        raise
                    else:
                        errors.append(e)
                else:
                    e.errors = tuple(errors)
                    raise

            else:
                # connection didn't fail, confirm it's live status
                self.connection_pool.mark_live(connection)

                if method == "HEAD":
                    return Response(
                        status=status,
                        headers=headers_response,
                        body=200 <= status < 300,
                    )

                if data:
                    data = self.deserializer.loads(
                        data, headers_response.get("content-type")
                    )

                # After the body is deserialized put the data
                # into one of the typed responses
                response_cls = Response
                if isinstance(data, list):
                    response_cls = ListResponse
                elif isinstance(data, dict):
                    response_cls = DictResponse
                return response_cls(
                    status=status,
                    headers=headers_response,
                    body=data,
                )

    def close(self):
        """
        Explicitly closes connections
        """
        self.connection_pool.close()


def _normalize_hosts(hosts, default_hosts):
    # if hosts are empty, just defer to defaults
    if hosts is None:
        return default_hosts or [{}]

    # passed in just one string
    if isinstance(hosts, string_types):
        hosts = [hosts]

    out = []
    # normalize hosts to dicts
    for host in hosts:
        if isinstance(host, string_types):
            if "://" not in host:
                host = "//%s" % host

            parsed_url = urlparse(host)
            h = {"host": parsed_url.hostname}

            if parsed_url.port is not None:
                h["port"] = parsed_url.port

            if parsed_url.scheme == "https":
                h["port"] = parsed_url.port if parsed_url.port is not None else 443
                h["use_ssl"] = True

            if parsed_url.path and parsed_url.path != "/":
                h["url_prefix"] = parsed_url.path

            out.append(h)
        else:
            out.append(host)
    return out
